import torch

def pad_if_needed(x, n):
    if not (x.size(-1) % n == 0):
        pad = torch.zeros(len(x.shape), dtype=torch.int)
        pad[-1] = n - (x.size(-1) % n)
        pad = tuple(pad.numpy().tolist())
        x = torch.nn.functional.pad(x, pad)
    return x

class GroupSum(torch.nn.Module):
    def __init__(self, k, tau=1, randperm=False):
        super().__init__()
        self.k = k
        self.tau = tau
        self.randperm = torch.randperm(randperm) if randperm else None

    def forward(self, x):
        if self.randperm is not None:
            x = x[:, self.randperm]
        x = pad_if_needed(x, self.k)
        x = x.view(*x.shape[:-1], self.k, int(x.shape[-1]/self.k))
        return x.sum(dim=-1) / self.tau

class STEFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        return (x > 0).float()

    @staticmethod
    def backward(ctx, output_grad):
        return output_grad # torch.nn.functional.hardtanh(output_grad)

class STE(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return STEFunction.apply(x)


import os
import importlib

dir_path = os.path.dirname(os.path.realpath(__file__))
modules = [f for f in os.listdir(dir_path) if f.endswith('.py') and f != '__init__.py']

for module in modules:
    module_name = module[:-3]
    submodule = importlib.import_module('.' + module_name, __name__)
    for item in dir(submodule):
        if not item.startswith("_"): 
            globals()[item] = getattr(submodule, item)



import torch
from torch.nn.functional import softmax

def layer_mapping(input_size, tuple_lenght, output_size, random=True):
    mapp = lambda : torch.randperm(input_size) if random else torch.arange(input_size)
    num_complete, offset = divmod(tuple_lenght*output_size, input_size)
    if num_complete == 0:
        return mapp()[:offset].reshape(output_size, tuple_lenght).type(torch.int32)
    else:
        complete_mappings = torch.cat([mapp() for i in range(num_complete)])
        offset_mapping = mapp()[:offset]
        return torch.cat([complete_mappings, offset_mapping]).reshape(output_size, tuple_lenght).type(torch.int32)

class LearnableMappingFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, weights, tau):
        mapping = weights.argmax(dim=0)
        output = x[:, mapping]
        ctx.save_for_backward(x, weights, tau)
        return output

    @staticmethod
    def backward(ctx, output_grad):
        x, weights, tau  = ctx.saved_tensors
        weights_grad = ((2*x-1).T @ output_grad)
        input_grad = output_grad @ softmax(weights/tau, dim=0).T
        return input_grad, weights_grad, None, None
    

class LearnableMapping(torch.nn.Module):
    def __init__(self, input_size, output_size, tau=0.001):
        super().__init__()
        self.weights = torch.nn.Parameter(torch.rand(input_size, output_size, dtype=torch.float32), requires_grad=True)
        self.tau = tau

    def forward(self, x):
        return LearnableMappingFunction.apply(x, self.weights, torch.tensor(self.tau))


import torch
if torch.cuda.is_available():
    import efd_cuda
from .mapping import layer_mapping, LearnableMapping
from .utils import STEFunction

class EFDFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, mapping, luts, alpha, beta):
        if x.is_cuda:
            output = efd_cuda.forward(x, mapping, luts)
        else:
            raise "EFDFunction CPU not Implemented"
        ctx.save_for_backward(x, mapping, luts, alpha, beta)
        return output

    @staticmethod
    def backward(ctx, output_grad):
        if output_grad.is_cuda:
            input_grad, luts_grad = efd_cuda.backward(*ctx.saved_tensors, output_grad.contiguous())
        else:
            raise "EFDFunction CPU not Implemented"
        return input_grad, None, luts_grad, None, None

class LUTLayer(torch.nn.Module):
    def __init__(self, input_size, output_size, n, mapping='random', alpha=None, beta=None, ste=True, 
                 clamp_luts=True, lm_tau=0.001):
        super().__init__()
        
        # Input Check
        assert input_size > 0
        assert output_size > 0
        assert n > 0
        assert mapping in ('arange', 'random', 'learnable') or (isinstance(mapping, torch.Tensor) and mapping.dtype == torch.int32 and mapping.shape == torch.Size([output_size, n]))
        assert isinstance(ste ,bool)
        assert isinstance(clamp_luts ,bool)

        # Vars
        self.input_size = int(input_size)
        self.output_size = int(output_size)
        self.n = int(n)
        self.ste = ste
        self.clamp_luts = clamp_luts

        # Alpha and beta
        if alpha is None:
            alpha = 0.5 * 0.75**(n-1)
        if beta is None:
            beta = 0.25/0.75
        self.alpha = torch.tensor(alpha)
        self.beta = torch.tensor(beta)
        assert self.alpha.dtype in (torch.float16, torch.float32, torch.float64)
        assert self.beta.dtype in (torch.float16, torch.float32, torch.float64)

        # Mapping
        if isinstance(mapping, torch.Tensor):
            self.mapping = torch.nn.Parameter(mapping, requires_grad=False)
        elif mapping == 'learnable':
            self.mapping = LearnableMapping(input_size, output_size*n, tau=lm_tau)
            self.__dummy_mapping = torch.nn.Parameter(torch.arange(output_size*n).reshape(output_size, n).int(), requires_grad=False)
        else:
            self.mapping = torch.nn.Parameter(layer_mapping(input_size, n, output_size, random=(mapping=='random')), requires_grad=False)
        
        # LUTs
        luts = torch.rand(output_size, 2**n, dtype=torch.float32)*2 - 1
        self.luts = torch.nn.Parameter(luts, requires_grad=True)
        self.luts.is_lut = True # Set for HammingL2 Reguralization
        
    def forward(self, x):
        
        # Clamp LUTs
        if self.training and self.clamp_luts:
            with torch.no_grad():
                self.luts.clamp_(-1, 1)
        
        # Learnable Mapping
        if isinstance(self.mapping, LearnableMapping):
            x = self.mapping(x)
            mapping = self.__dummy_mapping
        else:
            mapping = self.mapping
        
        # EFD
        x = EFDFunction.apply(x, mapping, self.luts, self.alpha, self.beta)

        # STE
        if self.ste:
            x = STEFunction.apply(x)
        
        return x


import os
import torch
import setuptools
from setuptools import setup
from torch.utils.cpp_extension import BuildExtension, CUDAExtension, CppExtension

cpp_dir = os.path.join('src', 'torch_dwn', 'custom_operators', 'cpp')
cuda_dir = os.path.join('src', 'torch_dwn', 'custom_operators', 'cuda')

ext_modules = []

for filename in os.listdir(cpp_dir):
    if filename[-3:] == 'cpp':
        ext_modules.append(CppExtension(filename[:-3], [os.path.join(cpp_dir, filename)]))

for filename in os.listdir(cuda_dir):
    if filename[-3:] == 'cpp':
        module_name = filename[:-4]
        kernel_filename = module_name + '_kernel.cu'
        ext_modules.append(CUDAExtension(module_name, [os.path.join(cuda_dir, filename), os.path.join(cuda_dir, kernel_filename)]))

setup(
    name='torch_dwn',
    ext_modules=ext_modules,
    cmdclass={'build_ext': BuildExtension},
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    version="1.0.9",
    author="Alan T. L. Bacellar",
    author_email="alanbacellar@gmail.com",
    description="Differentiable Weightless Neural Networks (DWN) PyTorch Module",
    url="https://github.com/alanbacellar/DWN",
    install_requires=[
        'torch'
    ]
)


import torch

class Thermometer:
    def __init__(self, num_bits=1, feature_wise=True):
        
        assert num_bits > 0
        assert type(feature_wise) is bool

        self.num_bits = int(num_bits)
        self.feature_wise = feature_wise
        self.thresholds = None

    def get_thresholds(self, x):
        min_value = x.min(dim=0)[0] if self.feature_wise else x.min()
        max_value = x.max(dim=0)[0] if self.feature_wise else x.max()
        return min_value.unsqueeze(-1) + torch.arange(1, self.num_bits+1).unsqueeze(0) * ((max_value - min_value) / (self.num_bits + 1)).unsqueeze(-1)

    def fit(self, x):
        if type(x) is not torch.Tensor:
            x = torch.tensor(x)
        self.thresholds = self.get_thresholds(x)
        return self
    
    def binarize(self, x):
        if self.thresholds is None:
            raise 'need to fit before calling apply'
        if type(x) is not torch.Tensor:
            x = torch.tensor(x)
        x = x.unsqueeze(-1)
        return (x > self.thresholds).float()

class GaussianThermometer(Thermometer):
    def __init__(self, num_bits=1, feature_wise=True):
        super().__init__(num_bits, feature_wise)

    def get_thresholds(self, x):
        std_skews = torch.distributions.Normal(0, 1).icdf(torch.arange(1, self.num_bits+1)/(self.num_bits+1))
        mean = x.mean(dim=0) if self.feature_wise else x.mean()
        std = x.std(dim=0) if self.feature_wise else x.std() 
        thresholds = torch.stack([std_skew * std + mean for std_skew in std_skews], dim=-1)
        return thresholds
    
class DistributiveThermometer(Thermometer):
    def __init__(self, num_bits=1, feature_wise=True):
        super().__init__(num_bits, feature_wise)

    def get_thresholds(self, x):
        data = torch.sort(x.flatten())[0] if not self.feature_wise else torch.sort(x, dim=0)[0]
        indicies = torch.tensor([int(data.shape[0]*i/(self.num_bits+1)) for i in range(1, self.num_bits+1)])
        thresholds = data[indicies]
        return torch.permute(thresholds, (*list(range(1, thresholds.ndim)), 0))

